



smu -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\include\\libsmu\\libsmu.hpp exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\bin\\libsmu.dll exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\lib\\libsmu.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
